import os, sys, inspect, platform

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)
sys.path.insert(0, parentdir+"/net_emulator")
# print(sys.path)

from objects.emulator import SimpleEmulator
from objects.cc_base import CongestionControl
from player.block_selection import Solution as BlockSelection
from net_utils.utils import *
from config.constant import *
from config import constant

from player.examples.reno import Reno
from double_flow import create_2flow_emulator
from qoe_model import cal_qoe


__all__ = ["SimpleEmulator", "CongestionControl", "BlockSelection", \
           "analyze_emulator", "plot_cwnd", "plot_rate", \
           "Reno", "create_2flow_emulator", "constant", \
           "cal_qoe"]

try:
    if os.path.exists("output"):
        if platform.system() == "Windows":
            # for windows
            os.system("rmdir /Q /S output")
        else:
            # for linux
            os.system("rm -rf output")

    # os.rmdir("output")
    os.mkdir("output")
    os.mkdir("output/packet_log")

except Exception as e:
    pass



# Copyright 2019 Nathan Jay and Noga Rotman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


# The monitor interval class used to pass data from the PCC subsystem to
# the machine learning module.
#
class SenderMonitorInterval():
    def __init__(self,
                 sender_id,
                 bytes_sent=0.0,
                 bytes_acked=0.0,
                 bytes_lost=0.0,
                 send_start=0.0,
                 send_end=0.0,
                 recv_start=0.0,
                 recv_end=0.0,
                 rtt_samples=[],
                 packet_size=1500):
        self.features = {}
        self.sender_id = sender_id
        self.bytes_acked = bytes_acked
        self.bytes_sent = bytes_sent
        self.bytes_lost = bytes_lost
        self.send_start = send_start
        self.send_end = send_end
        self.recv_start = recv_start
        self.recv_end = recv_end
        self.rtt_samples = rtt_samples
        self.packet_size = packet_size

    def get(self, feature):
        if feature in self.features.keys():
            return self.features[feature]
        else:
            result = SenderMonitorIntervalMetric.eval_by_name(feature, self)
            self.features[feature] = result
            return result

    # Convert the observation parts of the monitor interval into a numpy array
    def as_array(self, features):
        return np.array([self.get(f) / SenderMonitorIntervalMetric.get_by_name(f).scale for f in features])


class SenderHistory():
    def __init__(self, length, features, sender_id):
        self.features = features
        self.values = []
        self.sender_id = sender_id
        for i in range(0, length):
            self.values.append(SenderMonitorInterval(self.sender_id))

    def step(self, new_mi):
        self.values.pop(0)
        self.values.append(new_mi)

    def as_array(self):
        arrays = []
        for mi in self.values:
            arrays.append(mi.as_array(self.features))
        arrays = np.array(arrays).flatten()
        return arrays


class SenderMonitorIntervalMetric():
    _all_metrics = {}

    def __init__(self, name, func, min_val, max_val, scale=1.0):
        self.name = name
        self.func = func
        self.min_val = min_val
        self.max_val = max_val
        self.scale = scale
        SenderMonitorIntervalMetric._all_metrics[name] = self

    def eval(self, mi):
        return self.func(mi)

    def eval_by_name(name, mi):
        return SenderMonitorIntervalMetric._all_metrics[name].eval(mi)

    def get_by_name(name):
        return SenderMonitorIntervalMetric._all_metrics[name]


def get_min_obs_vector(feature_names):
    print("Getting min obs for %s" % feature_names)
    result = []
    for feature_name in feature_names:
        feature = SenderMonitorIntervalMetric.get_by_name(feature_name)
        result.append(feature.min_val)
    return np.array(result)


def get_max_obs_vector(feature_names):
    result = []
    for feature_name in feature_names:
        feature = SenderMonitorIntervalMetric.get_by_name(feature_name)
        result.append(feature.max_val)
    return np.array(result)


def _mi_metric_recv_rate(mi):
    dur = mi.get("recv dur")
    if dur > 0.0:
        return 8.0 * (mi.bytes_acked - mi.packet_size) / dur
    return 0.0


def _mi_metric_recv_dur(mi):
    return mi.recv_end - mi.recv_start


def _mi_metric_avg_latency(mi):
    if len(mi.rtt_samples) > 0:
        return np.mean(mi.rtt_samples)
    return 0.0


def _mi_metric_send_rate(mi):
    dur = mi.get("send dur")
    if dur > 0.0:
        return 8.0 * mi.bytes_sent / dur
    return 0.0


def _mi_metric_send_dur(mi):
    return mi.send_end - mi.send_start


def _mi_metric_loss_ratio(mi):
    if mi.bytes_lost + mi.bytes_acked > 0:
        return mi.bytes_lost / (mi.bytes_lost + mi.bytes_acked)
    return 0.0


def _mi_metric_latency_increase(mi):
    half = int(len(mi.rtt_samples) / 2)
    if half >= 1:
        return np.mean(mi.rtt_samples[half:]) - np.mean(mi.rtt_samples[:half])
    return 0.0


def _mi_metric_ack_latency_inflation(mi):
    dur = mi.get("recv dur")
    latency_increase = mi.get("latency increase")
    if dur > 0.0:
        return latency_increase / dur
    return 0.0


def _mi_metric_sent_latency_inflation(mi):
    dur = mi.get("send dur")
    latency_increase = mi.get("latency increase")
    if dur > 0.0:
        return latency_increase / dur
    return 0.0


_conn_min_latencies = {}


def _mi_metric_conn_min_latency(mi):
    latency = mi.get("avg latency")
    if mi.sender_id in _conn_min_latencies.keys():
        prev_min = _conn_min_latencies[mi.sender_id]
        if latency == 0.0:
            return prev_min
        else:
            if latency < prev_min:
                _conn_min_latencies[mi.sender_id] = latency
                return latency
            else:
                return prev_min
    else:
        if latency > 0.0:
            _conn_min_latencies[mi.sender_id] = latency
            return latency
        else:
            return 0.0


def _mi_metric_send_ratio(mi):
    thpt = mi.get("recv rate")
    send_rate = mi.get("send rate")
    if (thpt > 0.0) and (send_rate < 1000.0 * thpt):
        return send_rate / thpt
    return 1.0


def _mi_metric_latency_ratio(mi):
    min_lat = mi.get("conn min latency")
    cur_lat = mi.get("avg latency")
    if min_lat > 0.0:
        return cur_lat / min_lat
    return 1.0


SENDER_MI_METRICS = [
    SenderMonitorIntervalMetric("send rate", _mi_metric_send_rate, 0.0, 1e9, 1e7),
    SenderMonitorIntervalMetric("recv rate", _mi_metric_recv_rate, 0.0, 1e9, 1e7),
    SenderMonitorIntervalMetric("recv dur", _mi_metric_recv_dur, 0.0, 100.0),
    SenderMonitorIntervalMetric("send dur", _mi_metric_send_dur, 0.0, 100.0),
    SenderMonitorIntervalMetric("avg latency", _mi_metric_avg_latency, 0.0, 100.0),
    SenderMonitorIntervalMetric("loss ratio", _mi_metric_loss_ratio, 0.0, 1.0),
    SenderMonitorIntervalMetric("ack latency inflation", _mi_metric_ack_latency_inflation, -1.0, 10.0),
    SenderMonitorIntervalMetric("sent latency inflation", _mi_metric_sent_latency_inflation, -1.0, 10.0),
    SenderMonitorIntervalMetric("conn min latency", _mi_metric_conn_min_latency, 0.0, 100.0),
    SenderMonitorIntervalMetric("latency increase", _mi_metric_latency_increase, 0.0, 100.0),
    SenderMonitorIntervalMetric("latency ratio", _mi_metric_latency_ratio, 1.0, 10000.0),
    SenderMonitorIntervalMetric("send ratio", _mi_metric_send_ratio, 0.0, 1000.0)
]





import random
from config.constant import BYTES_PER_PACKET
import numpy as np


class Link():

    def __init__(self, trace_list, queue_size, **kwargs):
        """
        :param trace_list: [[time, bandwith, loss_rate, delay] ...]
        :param queue_size:
        """
        self.id = Link._get_next_id()
        self.trace_list = trace_list
        if len(trace_list) == 0:
            self.bandwith = np.inf if "bandwith" not in kwargs else kwargs["bandwith"]
            self.loss_rate = .0 if "loss_rate" not in  kwargs else kwargs["loss_rate"]
            self.delay = .001 if "delay" not in kwargs else kwargs["delay"]
        else:
            self.bandwith = trace_list[0][1] * 10**6 / BYTES_PER_PACKET
            self.loss_rate = trace_list[0][2]
            self.delay = trace_list[0][3]

        self.queue_delay = 0.0
        self.queue_delay_update_time = 0.0
        self.max_queue_delay = queue_size / self.bandwith

        self.extra_delay = 0
        self.queue_size = queue_size
        self.trace_start_id = 0

    _next_id = 1

    @classmethod
    def _get_next_id(cls):
        result = Link._next_id
        Link._next_id += 1
        return result

    def get_cur_queue_delay(self, event_time):
        return max(0.0, self.queue_delay - (event_time - self.queue_delay_update_time))

    def get_cur_latency(self, event_time):
        return self.delay + self.get_cur_queue_delay(event_time)

    def packet_enters_link(self, event_time):
        """check whether or not there is random loss or congestion loss in link at time of "event_time"."""
        if (random.random() < self.loss_rate):
            return False
        self.queue_delay = self.get_cur_queue_delay(event_time)
        self.queue_delay_update_time = event_time
        self.extra_delay = self.send_delay(event_time) # 1.0 / self.bandwith
        # print("Extra delay: %f, Current delay: %f, Max delay: %f" % (extra_delay, self.queue_delay, self.max_queue_delay))
        if self.extra_delay + self.queue_delay > self.max_queue_delay:
            # print("\tDrop!")
            return False
        self.queue_delay += self.extra_delay
        # print("\tNew delay = %f" % self.queue_delay)
        return True

    def send_delay(self, event_time):
        """calculate the time sent packet out of link when there is a changeable bandwidth, loss rate and delay."""
        rest_block_size = 1
        transmition_time = 0
        # different bandwith
        for i in range(self.trace_start_id, len(self.trace_list)):
            if rest_block_size <= 0:
                break
            if event_time + transmition_time > self.trace_list[i][0]:
                self.bandwith = self.trace_list[i][1] * 10 ** 6 / BYTES_PER_PACKET
                self.loss_rate = self.trace_list[i][2]
                self.delay = self.trace_list[i][3]
                continue

            used_time = rest_block_size / self.bandwith
            tmp = self.trace_list[i][0] - (event_time + transmition_time)
            if used_time > tmp:
                used_time = tmp
                rest_block_size -= used_time * self.bandwith
                self.bandwith = self.trace_list[i][1] * 10 ** 6 / BYTES_PER_PACKET
                self.loss_rate = self.trace_list[i][2]
                self.delay = self.trace_list[i][3]
            else:
                rest_block_size = 0
            transmition_time += used_time
            self.trace_start_id = i

        if rest_block_size > 0:
            transmition_time += rest_block_size / self.bandwith

        self.max_queue_delay = self.queue_size / self.bandwith

        return transmition_time

    def print_debug(self):
        print("Link: %d" % self.id)
        print("Bandwidth: %f" % self.bandwith)
        print("Delay: %f" % self.delay)
        print("Queue Delay: %f" % self.queue_delay)
        print("Max Queue Delay: %f" % self.max_queue_delay)
        print("One Packet Queue Delay: %f" % (1.0 / self.bandwith))

    def reset(self):
        self.queue_delay = 0.0
        self.queue_delay_update_time = 0.0


from objects.block import Block
from objects.packet import Packet
import numpy as np
import pandas as pd
from net_utils.utils import debug_print
import json


class AppicationLayer(object):

    def __init__(self,
                 block_file,
                 create_det=0.1,
                 bytes_per_packet=1500,
                 **kwargs):
        self.extra = {}
        self.update_config(kwargs)
        self.block_file = block_file
        self.block_queue = []
        self.bytes_per_packet = bytes_per_packet

        self.block_nums = None
        self.init_time = .0
        self.pass_time = .0
        self.fir_log = True

        self.now_block = None
        self.now_block_offset = 0
        self.head_per_packet = 20

        self.create_det = create_det
        self.handle_block(block_file)
        self.ack_blocks = dict()
        self.blocks_status = dict()

    def update_config(self, extra):
        self.extra["ENABLE_BLOCK_LOG"] = extra["ENABLE_BLOG_LOG"] if "ENABLE_BLOCK_LOG" in extra else True
        self.run_dir = extra["RUN_DIR"]+'/' if "RUN_DIR" in extra else ''

    def handle_block(self, block_file):
        """
        creating block queue by "block_file".
        :param block_file: str
        :return:
        """
        if isinstance(block_file, str):
            block_file = [block_file]
        for single_file in block_file:
            if single_file[-4:] == ".csv":
                self.create_blok_by_csv(single_file)
            else:
                self.create_block_by_file(single_file, self.create_det)
        self.block_queue = sorted(self.block_queue, key=lambda x:x.timestamp)

    def create_blok_by_csv(self, csv_file):
        df_data = pd.read_csv(csv_file, header=None)
        shape = df_data.shape
        assert len(shape) >= 2
        if shape[1] == 2:
            df_data.columns = ["time", "size"]
        elif shape[1]== 3:
            df_data.columns = ["time", "size", "key_frame"]

        for idx in range(shape[0]):
            det = 1
            priority = 0
            deadline = 0.2
            # change priority
            if "video" in csv_file:
                priority = 2
            elif "audio" in csv_file:
                priority = 1
            if "priority-" in csv_file:
                priority = int(csv_file[int(csv_file.index("priority-") + len("priority-"))])
            # change deadline
            if "ddl-" in csv_file:
                st = csv_file.index("ddl-") + len("ddl-")
                for i in range(st, len(csv_file)):
                    if csv_file[i] == '-':
                        deadline = float(csv_file[st:i])
                        break
            # create block
            block = Block(bytes_size=float(df_data["size"][idx])*det,
                          deadline=deadline,
                          priority=priority,
                          timestamp=float(df_data["time"][idx]))
            self.block_queue.append(block)

    def create_block_by_file(self, block_file, det=0.1):
        with open(block_file, "r") as f:
            self.block_nums = int(f.readline())

            pattern_cols = ["type", "size", "ddl"]
            pattern=[]
            for line in f.readlines():
                pattern.append(
                    { pattern_cols[idx]:item.strip() for idx, item in enumerate(line.split(',')) }
                )

            peroid = len(pattern)
            for idx in range(self.block_nums):
                ch = idx % peroid
                block = Block(bytes_size=float(pattern[ch]["size"]),
                              deadline=float(pattern[ch]["ddl"]),
                              timestamp=self.init_time+self.pass_time+idx*det,
                              priority=pattern[ch]["type"])
                self.block_queue.append(block)

    def select_block(self):
        """select the block that not sent and return it  """
        cur_time = self.init_time + self.pass_time
        # Is it necessary ? filter block with missing ddl
        for idx in range(len(self.block_queue)-1, -1, -1):
            item = self.block_queue[idx]
            # if miss ddl in queue, clean and log
            if cur_time > item.timestamp + item.deadline:
                self.block_queue[idx].miss_ddl = 1
                self.log_block(self.block_queue[idx])
                self.block_queue.pop(idx)
        
        if len(self.block_queue) == 0:
            return None
        best_block = self.block_queue.pop(0)

        return best_block

    def get_next_packet(self, cur_time, mode=None):
        """
        get the packet that can be sent at "cur_time" and return it to sender.
        :param cur_time: float, current time.
        :param mode: if mode is "force", it will return the packet after cur_time in case of lacking of packet.
        :return: Packet.
        """
        self.pass_time = cur_time
        if self.now_block is None:
            self.now_block = self.select_block()
            if self.now_block is None:
                return None
            self.ack_blocks[self.now_block.block_id] = []
            self.now_block_offset = 0
            self.now_block.split_nums = int(np.ceil(self.now_block.size /
                                            (self.bytes_per_packet - self.head_per_packet)))
            self.blocks_status[self.now_block.block_id] = self.now_block

        # It will only send the packet that already created if mode is None;
        # else will update system time
        if mode != "force" and cur_time < self.now_block.timestamp:
            return None
        payload = self.bytes_per_packet - self.head_per_packet
        if self.now_block.size % (self.bytes_per_packet - self.head_per_packet) and \
                self.now_block_offset == self.now_block.split_nums - 1:
            payload = self.now_block.size % (self.bytes_per_packet - self.head_per_packet)

        packet = Packet(create_time=self.now_block.timestamp,
                          next_hop=0,
                          offset=0,
                          packet_size=self.bytes_per_packet,
                          payload=payload,
                          block_info=self.now_block.get_block_info()
                          )
        self.now_block = None
        return packet

    def update_block_status(self, packet, cur_time):
        """update the block finishing status according to the acknowledge packets pushed from sender."""
        block_id = packet.block_info["Block_id"]
        self.pass_time = cur_time
        # filter repeating acked packet
        if block_id in self.ack_blocks and   \
                packet.offset in self.ack_blocks[block_id]:
            return

        # update block information.
        # Which is better? Save packet individual value or sum value
        self.blocks_status[block_id].send_delay += packet.send_delay
        # whether or not take pacing delay into consideration?
        self.blocks_status[block_id].latency += packet.latency
        self.blocks_status[block_id].finished_bytes += packet.payload

        if block_id not in self.ack_blocks:
            self.ack_blocks[block_id] = [packet.offset]
        # retransmission packet may be sended many times
        else:
            self.ack_blocks[block_id].append(packet.offset)

        if self.is_sent_block(block_id):
            self.blocks_status[block_id].finish_timestamp = packet.finish_time
            self.log_block(self.blocks_status[block_id])

    def log_block(self, block):
        """logging the finished blocks or the blocks with missing deadline"""
        if not self.extra["ENABLE_BLOCK_LOG"]:
            return
        if self.fir_log:
            self.fir_log = False
            with open(self.run_dir+"output/block.log", "w") as f:
                pass

        if not self.is_sent_block(block.block_id):
            block.finish_timestamp = self.init_time + self.pass_time
        if block.is_miss_ddl():
            block.miss_ddl = 1

        with open(self.run_dir+"output/block.log", "a") as f:
            f.write(json.dumps(block.trans2dict())+'\n')

    def is_sent_block(self, block_id):
        """check whether or not the block with the id of "block_id" is finished."""
        if block_id in self.ack_blocks and \
                len(self.ack_blocks[block_id]) == self.blocks_status[block_id].split_nums:
            return True
        return False

    def close(self, cur_time):
        """do some operations when system is closing, like logging the blocks with the packets that have not been acked or sent."""
        self.pass_time = cur_time
        for block_id, packet_list in self.ack_blocks.items():
            if self.is_sent_block(block_id):
                continue
            debug_print("block {} not finished!".format(block_id))
            self.log_block(self.blocks_status[block_id])
        return None



from net_utils.utils import measure_time


class Solution(object):

    # @measure_time()
    def select_block(self, cur_time, block_queue):
        '''
        The alogrithm to select the block which will be sended in next.
        The following example is selecting block by the create time firstly, and radio of rest life time to deadline secondly.
        :param cur_time: float
        :param block_queue: the list of Block.You can get more detail about Block in objects/block.py
        :return: int
        '''
        def is_better(block):
            best_block_create_time = best_block.block_info["Create_time"]
            cur_block_create_time = block.block_info["Create_time"]
            # if block is miss ddl
            if (cur_time - cur_block_create_time) >= block.block_info["Deadline"]:
                return False
            if (cur_time - best_block_create_time) >= best_block.block_info["Deadline"]:
                return True
            if best_block_create_time != cur_block_create_time:
                return best_block_create_time > cur_block_create_time
            return (cur_time - best_block_create_time) * best_block.block_info["Deadline"] > \
                   (cur_time - cur_block_create_time) * block.block_info["Deadline"]

        best_block_idx = -1
        best_block= None
        for idx, item in enumerate(block_queue):
            if best_block is None or is_better(item) :
                best_block_idx = idx
                best_block = item

        return best_block_idx



from config.constant import *
from common import sender_obs
from net_utils.utils import check_solution_format, measure_time
from objects.application import AppicationLayer
from objects.packet import Packet
from config import constant
import heapq


class Sender():

    def __init__(self, path, dest, features, history_len=10, solution=None):
        self.id = Sender._get_next_id()
        self.sent = 0
        self.acked = 0
        self.lost = 0
        self.bytes_in_flight = 0
        self.min_latency = None
        self.rtt_samples = []
        self.sample_time = []
        self.net = None
        self.path = path
        self.dest = dest
        self.history_len = history_len
        self.features = features
        self.history = sender_obs.SenderHistory(self.history_len,
                                                self.features, self.id)

        self.solution = solution
        ret = check_solution_format(self.solution.on_packet_sent(0))
        self.rate = ret["send_rate"] if "send_rate" in ret else float("inf")
        self.USE_CWND = self.solution.USE_CWND if hasattr(self.solution, "USE_CWND") else False
        # Not use this if USE_CWND=FALSE
        self.cwnd = ret["cwnd"] if "cwnd" in ret else 25
        self.starting_rate = self.rate
        self.pacing_rate = ret["pacing_rate"] if "pacing_rate" in ret else float("inf")
        self.cur_time = 0

        self.application = None
        # for cut the packet numbers that can't be sended due to cwnd in log
        self.wait_for_push_packets = []
        self.extra = {}
        # for player
        self.wait_for_select_packets = []
        self.in_event_nums = 0
        # record packet's decision order
        self.decision_order = 0
        # dict for packet needing to retransmission
        self.retrans_dict = dict()

    _next_id = 1

    @classmethod
    def _get_next_id(cls):
        result = Sender._next_id
        Sender._next_id += 1
        return result

    def init_application(self, block_file, **kwargs):
        """initial the sender's application which will send packet to it."""
        self.application = AppicationLayer(block_file, bytes_per_packet=BYTES_PER_PACKET, **kwargs)

    # @measure_time()
    def new_packet(self, cur_time, mode):
        """get new packet from it's application."""
        if self.application:
            packet = self.application.get_next_packet(cur_time, mode)
        else:
            packet = Packet.create_normal_packet(cur_time, packet_size=constant.BYTES_PER_PACKET)
        if packet:
            packet.send_delay = 1 / self.rate

        return packet

    def clear_miss_ddl(self, cur_time):
        """pop these packets with missing deadline at time of "cur_time"."""
        if not self.application:
            return
        for idx in range(len(self.wait_for_select_packets)-1, -1, -1):
            item = self.wait_for_select_packets[idx]
            if item.is_miss_ddl(cur_time):
                self.wait_for_select_packets.pop(idx)

    def append_retrans_packet(self):
        if self.application is None:
            return
        block_list = [packet.block_info["Block_id"] for packet in self.wait_for_select_packets]
        for block in self.retrans_dict:
            if block not in block_list and len(self.retrans_dict[block]) > 0:
                self.wait_for_select_packets.append(self.retrans_dict[block].pop(0))

    # @measure_time()
    def select_packet(self, cur_time):
        """
        select the packet that will be sent at time of "cur_time".
        Firstly, we will add all new packet from application that can be sent now.
        Secondly, if ENABLE_HASH_CHECK is True, we will do some safe check which will cost so much time.
        Thirdly, we will call your algorithm for packet selection and get the index that the packet will be sent now.
        :param cur_time:
        :return:
        """
        # append retrans packet which nat in selecing queue
        self.append_retrans_packet()
        # Is it necessary ? Reduce system burden by delete the packets missing ddl in time
        self.clear_miss_ddl(cur_time)
        while True:
            # if there is no packet can be sended, we need to send packet that created after cur_time
            packet = self.new_packet(cur_time, "force" if len(self.wait_for_select_packets) + len(self.wait_for_push_packets) == 0 else None)
            
            if not packet:
                break
            # if mode == force, update current time
            cur_time = max(cur_time, packet.create_time)
            
            self.wait_for_select_packets.append(packet)
            # for multi flow
            if self.application is None:
                return self.wait_for_select_packets.pop(0)
        
        if constant.ENABLE_HASH_CHECK:
            last_hash_vals = [item.get_hash_val() for item in self.wait_for_select_packets]
        packet_idx = self.solution.select_block(cur_time, self.wait_for_select_packets)
        # use hash for safety
        if constant.ENABLE_HASH_CHECK:
            now_hash_vals = [item.get_hash_val() for item in self.wait_for_select_packets]
            if last_hash_vals != now_hash_vals:
                raise ValueError("You shouldn't change the packet information in system!")
        if isinstance(packet_idx, int) and packet_idx >= 0:
            # set decision order in packet
            self.decision_order += 1
            _packet = self.wait_for_select_packets[packet_idx]
            _packet.decision_order = self.decision_order
            if self.application:
                # create next offset packet
                next_packet = _packet.next_offset()
                if next_packet:
                    self.wait_for_select_packets[packet_idx] = next_packet
                else:
                    block_id = _packet.block_info["Block_id"]
                    # whether or not there are rest of packet needing to retransmission
                    if block_id in self.retrans_dict and len(self.retrans_dict[block_id]) > 0:
                        self.wait_for_select_packets[packet_idx] = self.retrans_dict[block_id].pop(0)
                    else:
                        # finished block
                        self.wait_for_select_packets.pop(packet_idx)
            return _packet
        return None

    def apply_rate_delta(self, delta):
        delta *= DELTA_SCALE
        # print("Applying delta %f" % delta)
        if delta >= 0.0:
            self.set_rate(self.rate * (1.0 + delta))
        else:
            self.set_rate(self.rate / (1.0 - delta))

    def apply_cwnd_delta(self, delta):
        delta *= DELTA_SCALE
        # print("Applying delta %f" % delta)
        if delta >= 0.0:
            self.set_cwnd(self.cwnd * (1.0 + delta))
        else:
            self.set_cwnd(self.cwnd / (1.0 - delta))

    def can_send_packet(self, cur_time):
        """
        check it that can send packet now if your congestion control is based windows.
        Firstly, we will call the funtion of "on_packet_sent" in your solution and get rate, cwnd, extra.
        Then, we will update sender according to your parameters.
        :return: Boolean.
        """
        ret = self.solution.on_packet_sent(cur_time)
        self.rate = ret["send_rate"] if "send_rate" in ret else self.rate
        self.cwnd = ret["cwnd"] if "cwnd" in ret else self.cwnd
        self.pacing_rate = ret["pacing_rate"] if "pacing_rate" in ret else self.pacing_rate
        self.extra = ret["extra"] if "extra" in ret else {}
        if self.USE_CWND:
            return int(self.bytes_in_flight) / BYTES_PER_PACKET < self.cwnd
        else:
            return True

    def register_network(self, net):
        self.net = net

    def on_packet_sent(self, cur_time):
        """
        some operation when packet is sent successfully out of sender.
        Like updating the inflight numbers and calculation pacing delay.
        :param cur_time: type float.
        :return: pacing delay and extra information that can be attached to packet.
        """
        self.sent += 1
        self.bytes_in_flight += BYTES_PER_PACKET

        # the old time will <= cur_time if there is no pacing due to the sequential processing
        old_time = self.cur_time
        self.cur_time = max(cur_time, self.cur_time) + 1 / self.pacing_rate
        return max(old_time-cur_time, .0), self.extra

    def on_packet_acked(self, rtt, packet, event_time):
        """
        some operation when packet acknowledged successfully at sender.
        Like updating the numbers of acked and inflight and push the acked information to it's application.
        :param rtt: round-trip-time
        :param packet: type Packet.
        """
        self.acked += 1
        self.rtt_samples.append(rtt)
        if (self.min_latency is None) or (rtt < self.min_latency):
            self.min_latency = rtt
        self.bytes_in_flight -= BYTES_PER_PACKET
        if self.application:
            self.application.update_block_status(packet, event_time)

    def on_packet_lost(self, event_time, packet):
        """
        some operation when packet lost at sender.
        Like updating the numbers of lost and inflight and do the retranssmition.
        :param event_time: type float.
        :param packet:
        """
        self.lost += 1
        self.bytes_in_flight -= BYTES_PER_PACKET
        if self.application:
            # do retrans if lost
            retrans_packet = packet.create_retrans_packet(event_time)
            retrans_block_id = retrans_packet.block_info["Block_id"]
            # save retransmission packet in dict
            if retrans_block_id in self.retrans_dict:
                self.retrans_dict[retrans_block_id].append(retrans_packet)
            else:
                self.retrans_dict[retrans_block_id] = [retrans_packet]

    def slide_windows(self, cur_time, queue_size):
        ret = []
        for i in range(int(self.cwnd) - queue_size):
            if len(self.wait_for_push_packets) == 0:
                _packet = self.select_packet(cur_time + (1.0 / self.rate))
                if _packet is None:
                    return ret
            else:
                _packet = heapq.heappop(self.wait_for_push_packets)[2]
            ret.append(_packet)

        return ret

    def set_rate(self, new_rate):
        self.rate = new_rate
        # print("Attempt to set new rate to %f (min %f, max %f)" % (new_rate, MIN_RATE, MAX_RATE))
        if self.rate > MAX_RATE:
            self.rate = MAX_RATE
        if self.rate < MIN_RATE:
            self.rate = MIN_RATE

    def set_cwnd(self, new_cwnd):
        self.cwnd = int(new_cwnd)
        # print("Attempt to set new rate to %f (min %f, max %f)" % (new_rate, MIN_RATE, MAX_RATE))
        if self.cwnd > MAX_CWND:
            self.cwnd = MAX_CWND
        if self.cwnd < MIN_CWND:
            self.cwnd = MIN_CWND

    def get_waiting_ack_nums(self):
        """get the numbers of inflight."""
        return int(self.bytes_in_flight) // BYTES_PER_PACKET

    def record_run(self):
        smi = self.get_run_data()
        self.history.step(smi)

    def get_obs(self):
        return self.history.as_array()

    def get_run_data(self):
        obs_end_time = self.net.get_cur_time()

        # obs_dur = obs_end_time - self.obs_start_time
        # print("Got %d acks in %f seconds" % (self.acked, obs_dur))
        # print("Sent %d packets in %f seconds" % (self.sent, obs_dur))
        # print("self.rate = %f" % self.rate)
        # print(self.rtt_samples)
        return sender_obs.SenderMonitorInterval(
            self.id,
            bytes_sent=self.sent * BYTES_PER_PACKET,
            bytes_acked=self.acked * BYTES_PER_PACKET,
            bytes_lost=self.lost * BYTES_PER_PACKET,
            send_start=self.obs_start_time,
            send_end=obs_end_time,
            recv_start=self.obs_start_time,
            recv_end=obs_end_time,
            rtt_samples=self.rtt_samples,
            packet_size=BYTES_PER_PACKET
        )

    def reset_obs(self):
        self.sent = 0
        self.acked = 0
        self.lost = 0
        self.rtt_samples = []
        self.obs_start_time = self.net.get_cur_time()

    def print_debug(self):
        print("Sender: %d" % (self.id))
        # print("Obs: %s" % str(self.get_obs()))
        print("Rate: %f" % self.rate)
        print("Sent: %d" % self.sent)
        print("Acked: %d" % self.acked)
        print("Lost: %d" % self.lost)
        print("Min Latency: %s" % str(self.min_latency))

    def reset(self):
        # print("Resetting sender!")
        self.rate = self.starting_rate
        self.bytes_in_flight = 0
        self.min_latency = None
        self.reset_obs()
        self.history = sender_obs.SenderHistory(self.history_len,
                                                self.features, self.id)

    def __lt__(self, other):
        return False


from config.constant import *
from net_utils.utils import debug_print
from objects.cc_base import CongestionControl


class MTR(CongestionControl):

    def __init__(self):
        super(MTR, self).__init__()
        self.trace_list = []
        self.USE_CWND = False

    def init_trace(self, trace_file):
        with open(trace_file, 'r') as f:
            for line in f.readlines():
                self.trace_list.append(list(
                    map(lambda x: float(x), line.split(","))
                ))
        self.send_rate = self.trace_list[0][1] * 10**6 / BYTES_PER_PACKET /2

    def update_trace(self, cur_time):
        for i in range(len(self.trace_list)):
            if self.trace_list[i][0] >= cur_time:
                break
            self.send_rate = self.trace_list[i][1] * 10**6 / BYTES_PER_PACKET /2

    def on_packet_sent(self, cur_time):
        self.update_trace(cur_time)
        output = {
            "send_rate" : self.send_rate
        }
        return output        

    def cc_trigger(self, cur_time, event_info):
        self._input_list.append([cur_time, event_info])

        if event_info["event_type"] != PACKET_TYPE_TEMP:
            self.update_trace(cur_time)
            return {
                "send_rate" : self.send_rate
            }
        return None




#!/usr/bin/python
# -*- coding: utf-8 -*-


import os, sys, inspect
import numpy as np
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)

import json, shutil


def cal_qoe(x=0, run_dir=None):
    block_data = []
    qoe = .0
    tmp = [3, 2, 1]
    if run_dir:
        if run_dir[-1] != '/':
            run_dir += '/'
    else:
        run_dir = './'
    with open(run_dir + "output/block.log", "r") as f:
        for line in f.readlines():
            data = json.loads(line.replace("'", '"'))
            # not finished
            if data["Miss_ddl"] == 0 and data["Size"] - data["Finished_bytes"] > 0.000001:
                continue
            block_data.append(data)
    for block in block_data:        
        if block["Miss_ddl"] == 0:
            priority = float(tmp[int(block['Priority'])] / 3)
            qoe += priority
        else:
            qoe -= x*priority
    return qoe

def aitrans_cal_qoe(x=0.9, run_dir=None):
    block_data = []
    urgency = []
    priorities = []
    qoe = 0
    tmp = [3, 2, 1]
    if run_dir:
        if run_dir[-1] != '/':
            run_dir += '/'
    else:
        run_dir = './'
    with open(run_dir + "output/block.log", "r") as f:
        for line in f.readlines():
            data = json.loads(line.replace("'", '"'))
            # not finished
            if data["Miss_ddl"] == 0 and data["Size"] - data["Finished_bytes"] > 0.000001:
                continue
            block_data.append(data)
    for block in block_data:
        priority = float(tmp[int(block['Priority'])] / 3)
        priorities.append(priority)
        if block["Miss_ddl"] == 0:
            urgency.append(1)
        else:
            urgency.append(0)
            priorities[-1] *= 0
    for i in range(len(urgency)):
        qoe += x * priorities[i] + (1 - x) * urgency[i]
    return qoe

